//R: here kernels for SETI MultiBeam OpenCL build will be located
// This version is using fraction(a * b * b) = fraction((_a + fa) (_b + fb) ^ 2) for enhancing the precision
#define M_2PI              6.28318530717958647692f //R: it's single precision constant here !
//R: single precision direct chirp (w/o angle range reduce) gives too low precision results to be used here (kernel deleted).
//R: port of original chirp2 from CUDA app gives increased number of inconclusives due to relatively low precision;
//R: it can be replaced with CHIRP3 mod that does more precise computations but suffers in speed on lesser ATi GPUs
//R: Also, it (CHIRP3) can't be used for FERMI due to bad precision on those chips.
//R: Doube precision chirp wins in speed for 
#if OCL_DOUBLE_CHIRP
#if defined(cl_khr_fp64)   //R:double precision chirp here
#pragma OPENCL EXTENSION cl_khr_fp64 : enable  
#endif
#if defined(cl_amd_fp64)
#pragma OPENCL EXTENSION cl_amd_fp64 : enable  
#endif

__kernel 
void CalcChirpData_kernel_dp_cl(const double chirp_rate, __global float4* cx_DataArray, __global float4* cx_ChirpDataArray,
								 const double recip_sample_rate) 
{  
		uint i= get_global_id(0);

		float4 cx=cx_DataArray[i];
		float4 chirp;
		float2 c, d;
		float2 angf;
		double2 time;
		time.x = (2*i)*recip_sample_rate;
		time.y= (2*i+1)*recip_sample_rate;
	    double2 ang  = chirp_rate*(time*time);       
		//ang  = ang-floor(ang);
		ang  = ang-trunc(ang);//R: trunc faster than floor for ATi
		angf.x =(float)ang.x;
		angf.y =(float)ang.y;
		angf*=M_2PI;
		//d=sincos(angf,&c);
		d=native_sin(angf);
		c=native_cos(angf);
			// real = cx.x * c - cx.y * d;
		chirp.xz = cx.xz*c.xy-cx.yw*d.xy;
			// imag = cx.x * d + cx.y * c;
		chirp.yw = cx.xz*d.xy+cx.yw*c.xy;
		cx_ChirpDataArray[i] = chirp;
}


#elif OCL_CHIRP3 //R: this new dechirp needs debugging still
float __fmul_rn(float a, float b){return a*b;}
float __fadd_rn(float a, float b){return a+b;}
float __float2int_rn(float a){return rint(a);}
float __int2float_rn(int a){return (float)a;}
float2 split(float a)
{
	const float split= 4097;
	float t = __fmul_rn(a,split);
	float2 r;
	r.x = t-(t-a);
	r.y = a - r.x;
	return r;
}
 float2 twoProd(float a, float b)
{
	float2 r;
	r.x = __fmul_rn(a,b);
	float2 aS = split(a);
	float2 bS = split(b);
	r.y = ( ( __fmul_rn(aS.x,bS.x) - r.x )
				+ __fmul_rn(aS.x,bS.y) + __fmul_rn(aS.y,bS.x) )
				+ __fmul_rn(aS.y,bS.y);
	return r;
}

float2 quickTwoSum(float a, float b)
{
	float2 r;
	r.x = a + b;
	r.y = b - (r.x-a);
	return r;
}

float2 df64_mult(float2 a, float2 b)
{
	float ah=a.x,al=a.y,bh=b.x,bl=b.y;
	float2 p; 
	p = twoProd(ah,bh);
	p.y += ah*bl;
	p.y += al*bh;
	p = quickTwoSum(p.x,p.y);
	return p;
}

__kernel
void CalcChirpData_kernel_df64_cl(const float2 chirp_rate, __global float4* cx_DataArray, 
							  __global float4* cx_ChirpDataArray, const float2 recip_sample_rate) {  
	uint i= get_global_id(0);   
//Jason: translating to intrinsics for IEEE compliance, better agreement with Fermi & CPU
//R:in OpenCL round to nearest is default rounding mode. intrinsics are emulated.     
		float4 cx = cx_DataArray[i];
		float2 c, d;
		float4 chirped;
		float4 time;
		time.xy=df64_mult(split(__int2float_rn(2*i)),recip_sample_rate);
		time.zw=df64_mult(split(__int2float_rn(2*i+1)),recip_sample_rate);
		float4 angdf;
		angdf.xy = df64_mult(chirp_rate,df64_mult(time.xy,time.xy));
		angdf.zw = df64_mult(chirp_rate,df64_mult(time.zw,time.zw));
		angdf = (angdf-rint(angdf))*M_2PI;
		float2 ang;
		ang.xy= angdf.xz+angdf.yw; 
		d=native_sin(ang);
		c=native_cos(ang);

//		real = cx.x * c - cx.y * d;
//		imag = cx.x * d + cx.y * c;
		chirped.xz = cx.xz*c - cx.yw*d;
		chirped.yw = cx.xz*d+cx.yw*c;

		cx_ChirpDataArray[i] = chirped;
}
#else 
__kernel __attribute__((vec_type_hint(float4)))
void CalcChirpData_kernel2_cl(const float4 chirp_rate, __global float4* cx_DataArray,
									   __global float4* cx_ChirpDataArray,__global float4* sample_rate) {  
	uint i= get_global_id(0);   
	float4 cx = cx_DataArray[i];
	float4 time = sample_rate[i];
	float sin1,sin2, cos1,cos2;
	float4 chirped;
	float tmp;
	float a = chirp_rate.x;
	float b = time.x;
	float a1= chirp_rate.z;
	float b1= time.z;
	float fa = chirp_rate.y;
	float fb = time.y;     
	float fa1 = chirp_rate.w;
	float fb1 = time.w;     
	float ang  = fract(2 * a * fract(b * fb,&tmp),&tmp) + fract(a * fb * fb,&tmp) + 
		fract(b * fract(b * fa,&tmp),&tmp) + fract(2 * b * fa * fb,&tmp) + fa * fb * fb;
	ang -= floor(ang);
	ang *= M_2PI;
	//sin1=sincos(ang,&cos1); 
	sin1=native_sin(ang);
	cos1=native_cos(ang);
	chirped.x = cx.x * cos1 - cx.y * sin1;
	chirped.y = cx.x * sin1 + cx.y * cos1;
	ang  = fract(2 * a1 * fract(b1 * fb1,&tmp),&tmp) + fract(a1 * fb1 * fb1,&tmp) + 
		fract(b1 * fract(b1 * fa1,&tmp),&tmp) + fract(2 * b1 * fa1 * fb1,&tmp) + fa1 * fb1 * fb1;
	ang -= floor(ang);
	ang *= M_2PI;
	//sin2=sincos(ang,&cos2); 
	sin2=native_sin(ang);
	cos2=native_cos(ang);
	chirped.z = cx.z * cos2 - cx.w * sin2;
	chirped.w = cx.z * sin2 + cx.w * cos2;

		cx_ChirpDataArray[i] = chirped;
}
#endif
#if USE_OPENCL_NV
__kernel void GetPowerSpectrum_kernel_cl(__global float2* FreqData, __global float* PowerSpectrum) {
	uint i= get_global_id(0);   
	float2 f1 = FreqData[i];
	float p;
	p=mad(f1.x,f1.x,f1.y*f1.y);
	PowerSpectrum[i] = p;
}
#else
__kernel void GetPowerSpectrum_kernel_cl(__global float4* FreqData, __global float2* PowerSpectrum) {
	uint i= get_global_id(0);   
	float4 f1 = FreqData[i];
	float2 p;
	p.xy=mad(f1.xz,f1.xz,f1.yw*f1.yw);
	PowerSpectrum[i] = p;
}
#endif
#if 0
typedef struct{
	float mean;
	float maximum;
	float index;
} SpikeResults;//R: there is no float3 in OpenCL 1.0 s using such structure
#endif
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_kernel_cl(__global float4* PowerSpectrum, __global float2* MeanMaxIdx, uint fftlen,
									  float best, float spike_thresh,__global uint* result_flag){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	//float mean=0.f;
	float4 sum4=0.f;
	float maxval=0.f;
	//float power;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
//	float4 mmi;
	float4 p;
	//R: separate handling of first quad - DC element should be omitted
	p=PowerSpectrum[tid*(fftlen>>2)];
	sum4.x=p.x;
	sum4.y=p.y;
	sum4.z=p.z;
	sum4.w=p.w;
	if(p.y>maxval){pos=1;}
	maxval=max(maxval,p.y);
	if(p.z>maxval){pos=2;}
	maxval=max(maxval,p.z);
	if(p.w>maxval){pos=3;}
	maxval=max(maxval,p.w);
	for(uint i=1;i<(fftlen>>2);i++){
		p=PowerSpectrum[i+tid*(fftlen>>2)];//R:doing 4 elements at once
		sum4+=p;
		if(p.x>maxval){pos=i*4;}
		maxval=max(maxval,p.x);
		if(p.y>maxval){pos=i*4+1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=i*4+2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=i*4+3;}
		maxval=max(maxval,p.w);
	}
	float2 res;//x-power, y - pos
	res.x=maxval*fftlen/(sum4.x+sum4.y+sum4.z+sum4.w);//R: mean is just the sum actually hence multiplication on fftlen
	res.y=pos;
	MeanMaxIdx[tid]=res;
//	MeanMaxIdx[3*tid+1]=maxval;
//	MeanMaxIdx[2*tid+1]=pos;
//	float power=maxval/mean;
//	float score=log10(power/40.f);
	if(res.x>best || res.x>spike_thresh){
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: cpu attention required to this block
	}
}
#if USE_OPENCL_NV || USE_OPENCL_HD5xxx
__kernel void PC_find_spike32_kernel_cl(__global float* PowerSpectrum, __global float2* MeanMaxIdx, uint Nchunk,
									  //float best, float spike_thresh,__global uint4* result_flag,
									  __local float4* tmp){
//R: float4 in shared mem is waste of resources, but CUDA SDK3.2RC2 doesn't understand float3... weird...    
	const int tid = get_local_id(0);
    const int y = get_global_id(1);
    const int gridX = get_local_size(0);
    const int start = y * gridX * Nchunk;
    const int end = (y + 1) * gridX*Nchunk;
	
    float sum = 0.0f;
    float maximum = 0.0f;
    int pos = 0;

	int i = start + tid;
	float val = PowerSpectrum[i]; 
	sum += val;//R: sum calculated over all elements while peak searched only through non-DC components
	if (get_global_id(0) > 0) { // Original max omits first element of every fft
		if (val > maximum)
			pos = i;
		maximum = max(maximum, val);
	}

    for (i = start + tid + gridX; i < end; i += gridX) {
        val = PowerSpectrum[i];
        sum += val;
        if (val > maximum)
            pos = i;
        maximum = max(maximum, val);
    }
	
    tmp[tid] =(float4)(sum, maximum, (float)(pos - start),0.f);

    for (i = gridX >> 1; i > 0; i >>= 1) {
        barrier(CLK_LOCAL_MEM_FENCE); 
        if (tid < i) {            
            tmp[tid] = (float4)(
                tmp[tid].x + tmp[tid + i].x, 
                max(tmp[tid].y, tmp[tid + i].y),
                tmp[tid].y > tmp[tid + i].y ? tmp[tid].z : tmp[tid + i].z,0.f//R:w not used
                ); 
        }
    }
    barrier(CLK_LOCAL_MEM_FENCE); 
    if (tid == 0) {
		//float4 mp4;
		//tmp4.x/=(gridX * Nchunk);//mean value needed, not just sum; sum/fftlen
		float2 power;
			power.x=(gridX * Nchunk)*tmp[0].y/tmp[0].x;
			power.y=tmp[0].z;
		//tmp4.xyzw=tmp[tid].xyzw;//R:float4 in output array but float3 in shared memory
		MeanMaxIdx[y]=power;
	//MeanMaxIdx[3*tid+2]=tmp4.z;
	//	float power=tmp4.y/tmp4.x;//max/mean
	//	float score=log10(power/40.f);
	//	if(score>best || power>spike_thresh){
	//		result_flag[0].x=1;//R: only in this case kernel results are needed
	//	}
	}

}
#endif
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_reduce0_kernel_cl(__global float4* PowerSpectrum, __global float4* tmp, uint fftlen){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	uint y=get_global_id(1);
	float4 sum4=(float4)0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
	float4 sum_val_pos;
	float4 p;
	int i=0;

	//R: separate handling of first quad - DC element should be omitted
	if(y==0){
		p=PowerSpectrum[tid*(fftlen>>2)];
		sum4.x=p.x;
		sum4.y=p.y;
		sum4.z=p.z;
		sum4.w=p.w;
		if(p.y>maxval){pos=1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=3;}
		maxval=max(maxval,p.w);
		i=1;
	}
	for(;i<(64);i++){
		p=PowerSpectrum[i+y*64+tid*(fftlen>>2)];//R:doing 4 elements at once
		sum4+=p;
		if(p.x>maxval){pos=i*4;}
		maxval=max(maxval,p.x);
		if(p.y>maxval){pos=i*4+1;}
		maxval=max(maxval,p.y);
		if(p.z>maxval){pos=i*4+2;}
		maxval=max(maxval,p.z);
		if(p.w>maxval){pos=i*4+3;}
		maxval=max(maxval,p.w);
	}
	sum_val_pos.x=sum4.x+sum4.y+sum4.z+sum4.w;
	sum_val_pos.y=maxval;
	sum_val_pos.z=pos+256*y;
	//R: w field not used
	tmp[tid*(fftlen>>8)+y]=sum_val_pos;
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_spike_reduce1_kernel_cl(__global float4* tmp, __global float* MeanMaxIdx, uint fftlen){
	//R: CUDA version of this kernel use shared memory and reducing, HD4xxx have no shared memory so I will
	//do whole power array per workitem.
	uint tid= get_global_id(0);
	//float mean=0.f;
	float sum=0.f;
	float maxval=0.f;
	float pos=0.f;//R: should be integer value but will use float for better memory packing
	//float4 mmi;
	float4 sum_val_pos;
	for(int y=0;y<(fftlen>>8);y++){
		sum_val_pos=tmp[y+tid*(fftlen>>8)];
		sum+=sum_val_pos.x;
		if(sum_val_pos.y>maxval){pos=sum_val_pos.z;}
		maxval=max(maxval,sum_val_pos.y);
	}
	float power=maxval*fftlen/sum;
	//mmi.x=mean;
	//mmi.y=maxval;
	//mmi.z=pos;
	//R: w field not used
	MeanMaxIdx[2*tid]=power;//mmi.x;
	MeanMaxIdx[2*tid+1]=pos;//mmi.y;
	//MeanMaxIdx[3*tid+2]=mmi.z;
	//float power=maxval/mean;
	//float score=log10(power/40);
	//if(score>best || power>spike_thresh){
	//	result_flag[0].x=1;//R: only in this case kernel results are needed
	//}
}


__kernel __attribute__((vec_type_hint(float4)))
void Transpose4_kernel_cl(__global float4* in, __global float4* out,uint x_size,uint y_size){
	//R: ATI HD4xxx has no cache exposed in OpenCL so do single possible memory optimization - vector reads in register file.
	//R: x_size matrix width, y_size - matrix heigh
	uint tidx= get_global_id(0);
	uint tidy= get_global_id(1);
	float4 d0=in[tidx+(x_size>>2)*(tidy<<2)];
	float4 d1=in[tidx+(x_size>>2)*((tidy<<2)+1)];
	float4 d2=in[tidx+(x_size>>2)*((tidy<<2)+2)];
	float4 d3=in[tidx+(x_size>>2)*((tidy<<2)+3)];
	float4 o0,o1,o2,o3;
	o0.x=d0.x;o0.y=d1.x;o0.z=d2.x;o0.w=d3.x;
	o1.x=d0.y;o1.y=d1.y;o1.z=d2.y;o1.w=d3.y;
	o2.x=d0.z;o2.y=d1.z;o2.z=d2.z;o2.w=d3.z;
	o3.x=d0.w;o3.y=d1.w;o3.z=d2.w;o3.w=d3.w;
	out[tidy+(y_size>>2)*(tidx<<2)]=o0;
	out[tidy+(y_size>>2)*((tidx<<2)+1)]=o1;
	out[tidy+(y_size>>2)*((tidx<<2)+2)]=o2;
	out[tidy+(y_size>>2)*((tidx<<2)+3)]=o3;
}

__kernel __attribute__((vec_type_hint(float4)))
void GetFixedPoT_kernel_cl(__global float4* power, __global float4* PoT,int fftlen) {
	int ul_PoT_i  = get_global_id(0);//R: dots position in single PoT array (0 to 64)
	int ul_PoT = get_global_id(1);//R: index of PoT array (there are fflen/4 (each workitem handles 1 dot for 4 PoT arrays) number of such arrays, each 64 points wide)

	int   
		ul_PoTChunkSize,
		ul_PoTChunk_i,
		ul_PoTChunkLimit,
		tPowerLen;//R: length of transposed power array

	tPowerLen = 1024*1024 / fftlen;//R: we have 1M of dots in initial power 2D matrix
	ul_PoTChunkSize  = tPowerLen / 64;//R: for now PoT array size for Gauss fitting is 64, this size in dots, not dots4

	// If the number of spectra is greater than the number
	// of elements in a PoT array, we add sum adjacent spectra
	// into PoT elements.
	// ul_PoTChunkSize indicates how many time-wise
	// power spectra bins are added together to make one PoT bin.

	ul_PoTChunk_i = ul_PoTChunkSize * ul_PoT_i;
	ul_PoTChunkLimit = ul_PoTChunk_i + ul_PoTChunkSize;

	float4 sum = (float4)0.0f;
	for (; ul_PoTChunk_i < ul_PoTChunkLimit; ul_PoTChunk_i++) {
		sum += power[(fftlen>>2)*ul_PoTChunk_i+ul_PoT];
	}
	PoT[ul_PoT_i*(fftlen>>2)+ul_PoT] = sum;
}

//R: TODO check if pre-check should be added in Normalize kernel to skip some areas of PoT matrix
__kernel __attribute__((vec_type_hint(float4)))
void NormalizePoT_kernel_cl(int ul_FftLength,__global float4* PoT,
									 __global float4* PoTPrefixSum, __global float4* NormMaxPower) {
	int tid= get_global_id(0);//R: should be from 0 to ul_FftLength/4 
//R: do 4 PoT arrays at once
//TODO: try to allocate private array of 64 float4 elements in register file to eliminate secondary read from
//global memory. But check of assembly output required to make sure private array not spilled into global memory
	float4 TotalPower = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		TotalPower += PoT[i * (ul_FftLength>>2)+tid];
	}
	float4 MeanPower = TotalPower / (float4)64;

	// Normalize power-of-time
	float4 sum = (float4)0.0f;
	float4 NormMaxPower_i = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		float4 PoTi = PoT[i * (ul_FftLength>>2)+tid] /MeanPower;
		NormMaxPower_i = max(NormMaxPower_i, PoTi);
		sum += PoTi;
		PoT[i * (ul_FftLength>>2)+tid] = PoTi;
		PoTPrefixSum[i * (ul_FftLength>>2)+tid] = sum;
	}
	NormMaxPower[tid] = NormMaxPower_i;
}



typedef struct {  
    int iSigma;
//    int gauss_pot_length;
    float GaussPowerThresh;
    float GaussPeakPowerThresh3;
    float GaussSigmaSq;
    int GaussTOffsetStart;
    int GaussTOffsetStop;
    float PeakScaleFactor;
    float GaussChiSqThresh;
    float gauss_null_chi_sq_thresh;
	float gauss_chi_sq_thresh;
//	float* dev_PoT;
//    float* dev_PoTPrefixSum;
//    float* dev_PowerSpectrum;
//    float4* dev_GaussFitResults;
//    float4* dev_GaussFitResultsReordered;
//    float4* dev_GaussFitResultsReordered2;
//    float* dev_NormMaxPower;    
//    float* dev_outputposition;
	float score_offset;
//    int NumDataPoints;
//	float f_weight[CUDA_ACC_MAX_GaussTOffsetStop]; // cached  static_cast<float>(EXP(i, 0, PoTInfo.GaussSigmaSq));
//    unsigned int nsamples;
} ocl_GaussFit_t;
__kernel __attribute__((vec_type_hint(float4)))
void NormalizePoT_peak_PC_kernel_cl(int ul_FftLength,__global float4* PoT,
									 __global float4* PoTPrefixSum, __global float4* NormMaxPower,
									 __constant ocl_GaussFit_t* settings,__global uint* WeakPeaks) {
	int tid= get_global_id(0);//R: should be from 0 to ul_FftLength/4 
//R: do 4 PoT arrays at once
//TODO: try to allocate private array of 64 float4 elements in register file to eliminate secondary read from
//global memory. But check of assembly output required to make sure private array not spilled into global memory
	float4 TotalPower = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		TotalPower += PoT[i * (ul_FftLength>>2)+tid];
	}
	float4 MeanPower = TotalPower / (float4)64;

	// Normalize power-of-time
	float4 sum = (float4)0.0f;
	float4 NormMaxPower_i = (float4)0.0f;
	for (int i = 0; i < 64; i++) {
		float4 PoTi = PoT[i * (ul_FftLength>>2)+tid] /MeanPower;
		NormMaxPower_i = max(NormMaxPower_i, PoTi);
		sum += PoTi;
		PoT[i * (ul_FftLength>>2)+tid] = PoTi;
		PoTPrefixSum[i * (ul_FftLength>>2)+tid] = sum;
	}
	NormMaxPower[tid] = NormMaxPower_i;
	float thresh=settings->GaussPowerThresh;
	if(NormMaxPower_i.x>thresh ||NormMaxPower_i.y>thresh ||
		NormMaxPower_i.z>thresh || NormMaxPower_i.w>thresh)	WeakPeaks[tid]=1;
	else WeakPeaks[tid]=0;
}


float gammln(float a) {
	float x,y,tmp,ser;
	float cof[6]={76.18009172947146f,-86.50532032941677f,
		24.01409824083091f,-1.231739572450155f,
		0.1208650973866179e-2f,-0.5395239384953e-5f};

	y=x=a;
	tmp=x+5.5f;
	tmp -= (x+0.5f)*log(tmp);
	ser=1.000000000190015f;
	for (int j=0;j<=5;j++){y+=1.f; ser += cof[j]/y;}
	return (float)(-tmp+log(2.5066282746310005f*ser/x));
}
#define ITMAX 10000  // Needs to be a few times the sqrt of the max. input to lcgf
//R: this function called only from device
//R: wlll process 4 PoTs at once as whole ATI OpenCL gauss search chain does - memory access optimization and better computing density
float4 GetTrueMean2(__global float4* fp_PoTPrefixSum, int ul_TOffset, int ul_ExcludeLen, int ul_FftLength,int tid) {
	// TrueMean is the mean power of the data set minus all power
	// out to ExcludeLen from our current TOffset.
	int i_start, i_lim;
	float4 f_ExcludePower;

	// take care that we do not add to exclude power beyond PoT bounds!
	i_start = max(ul_TOffset - ul_ExcludeLen, 0) - 1;
	i_lim = min(ul_TOffset + ul_ExcludeLen + 1, 64) - 1;

	f_ExcludePower = fp_PoTPrefixSum[i_lim * (ul_FftLength>>2)+tid];
	if (i_start >= 0)
		f_ExcludePower -= fp_PoTPrefixSum[i_start * (ul_FftLength>>2)+tid]; 

	return(((float4)(64.0f) - f_ExcludePower) / (float4)(64 - (i_lim - i_start)));
}
float4 GetPeak(__global float4* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float4 f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,int tid,__constant float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float4 f_sum=(float4)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i * (ul_FftLength>>2)+tid] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}

float4 GetChiSq(__global float4* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float4 f_PeakPower, float4 f_MeanPower, float4* xsq_null,int tid,__constant float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float4 f_ChiSq = (float4)0.0f, f_null_hyp=(float4)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float4 recip_MeanPower = (float4)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float4 f_PredictedPower = f_MeanPower + f_PeakPower * (float4)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float4 noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float4 recip_noise = (float4)rebin / noise;

		float4 PoTval = fp_PoT[i * (ul_FftLength>>2)+tid] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float4)(64.0f);
	f_null_hyp/=(float4)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}
float4 lcgf(float a, float4 x) {
	const float EPS= 1.19209e-006f; //007;//std::numeric_limits<double>::epsilon();
	const float FPMIN= 9.86076e-031f; //032;//std::numeric_limits<double>::min()/EPS;
	float4 an,b,c,d,del,h;
	//float gln=gammln(a);
	//R: only 31 or 31.5 values for a so I put precomputed gammln(a) here.
	float gln=(a==31.f)?74.658233642578125f:76.371200561523437f;

	b=x+(float4)(1.0f-a);
	c=(float4)(1.0f/FPMIN);
	d=(float4)1.0f/b;
	h=d;
	for (int i=1;i<=ITMAX;++i) {
		an = -i*((float)i-a);
		b += (float4)2.0f;
		d=an*d+b;
		if (fabs(d.x)<FPMIN) d.x=FPMIN;
		if (fabs(d.y)<FPMIN) d.y=FPMIN;
		if (fabs(d.z)<FPMIN) d.z=FPMIN;
		if (fabs(d.w)<FPMIN) d.w=FPMIN;
		c=b+an/c;
		if (fabs(c.x)<FPMIN) c.x=FPMIN;
		if (fabs(c.y)<FPMIN) c.y=FPMIN;
		if (fabs(c.z)<FPMIN) c.z=FPMIN;
		if (fabs(c.w)<FPMIN) c.w=FPMIN;
		d=(float4)1.0f/d;
		del=d*c;
		h*=del;
		if (fabs(del.x-1.0f)<EPS && fabs(del.y-1.0f)<EPS && fabs(del.z-1.0f)<EPS && fabs(del.w-1.0f)<EPS) break;
	}

	return (log(h)-x+a*log(x)-(float4)gln);
}

float4 calc_GaussFit_score(float4 chisqr, float4 null_chisqr,float score_offset) {
	float gauss_bins = 64.0f;
	float gauss_dof = (gauss_bins-2.0f) * 0.5f;
	float null_dof = (gauss_bins-1.0f) * 0.5f;
	gauss_bins *= 0.5f;
	return  score_offset +
		lcgf(gauss_dof,max(chisqr*gauss_bins,(float4)(gauss_dof+1.0f)))
		//-cudaAcc_lcgf(gauss_dof,cudaAcc_GaussFit_settings.gauss_chi_sq_thresh*gauss_bins) // <- always the same result
		-lcgf(null_dof,max(null_chisqr*gauss_bins,null_dof+1.0f));
		//+cudaAcc_lcgf(null_dof,cudaAcc_GaussFit_settings.gauss_null_chi_sq_thresh*gauss_bins); // <- always the same result	
}
float GetTrueMean2_fl(__global float* fp_PoTPrefixSum, int ul_TOffset, int ul_ExcludeLen, int ul_FftLength,int tid) {
	// TrueMean is the mean power of the data set minus all power
	// out to ExcludeLen from our current TOffset.
	int i_start, i_lim;
	float f_ExcludePower;

	// take care that we do not add to exclude power beyond PoT bounds!
	i_start = max(ul_TOffset - ul_ExcludeLen, 0) - 1;
	i_lim = min(ul_TOffset + ul_ExcludeLen + 1, 64) - 1;

	f_ExcludePower = fp_PoTPrefixSum[i_lim * (ul_FftLength)+tid];
	if (i_start >= 0)
		f_ExcludePower -= fp_PoTPrefixSum[i_start * (ul_FftLength)+tid]; 

	return(((float)(64.0f) - f_ExcludePower) / (float)(64 - (i_lim - i_start)));
}
float GetPeak_fl(__global float* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,int tid,__constant float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float f_sum=(float)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i * (ul_FftLength)+tid] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}

float GetChiSq_fl(__global float* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float f_PeakPower, float f_MeanPower, float* xsq_null,int tid,__constant float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float f_ChiSq = (float)0.0f, f_null_hyp=(float)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float recip_MeanPower = (float)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float f_PredictedPower = f_MeanPower + f_PeakPower * (float)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float recip_noise = (float)rebin / noise;

		float PoTval = fp_PoT[i * (ul_FftLength)+tid] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float)(64.0f);
	f_null_hyp/=(float)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}
float lcgf_fl(float a, float x) {
	const float EPS= 1.19209e-006f; //007;//std::numeric_limits<double>::epsilon();
	const float FPMIN= 9.86076e-031f; //032;//std::numeric_limits<double>::min()/EPS;
	float an,b,c,d,del,h;
	//float gln=gammln(a);
	//R: only 31 or 31.5 values for a so I put precomputed gammln(a) here.
	float gln=(a==31.f)?74.658233642578125f:76.371200561523437f;

	b=x+(float)(1.0f-a);
	c=(float)(1.0f/FPMIN);
	d=(float)1.0f/b;
	h=d;
	for (int i=1;i<=ITMAX;++i) {
		an = -i*((float)i-a);
		b += (float)2.0f;
		d=an*d+b;
		if (fabs(d)<FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c)<FPMIN) c=FPMIN;
		d=(float)1.0f/d;
		del=d*c;
		h*=del;
		if (fabs(del-1.0f)<EPS) break;
	}

	return (log(h)-x+a*log(x)-(float)gln);
}

float calc_GaussFit_score_fl(float chisqr, float null_chisqr,float score_offset) {
	float gauss_bins = 64.0f;
	float gauss_dof = (gauss_bins-2.0f) * 0.5f;
	float null_dof = (gauss_bins-1.0f) * 0.5f;
	gauss_bins *= 0.5f;
	return  score_offset +
		lcgf_fl(gauss_dof,max(chisqr*gauss_bins,(float)(gauss_dof+1.0f)))
		//-cudaAcc_lcgf(gauss_dof,cudaAcc_GaussFit_settings.gauss_chi_sq_thresh*gauss_bins) // <- always the same result
		-lcgf_fl(null_dof,max(null_chisqr*gauss_bins,null_dof+1.0f));
		//+cudaAcc_lcgf(null_dof,cudaAcc_GaussFit_settings.gauss_null_chi_sq_thresh*gauss_bins); // <- always the same result	
}

#if USE_OPENCL_NV || USE_OPENCL_HD5xxx //R: this kernel uses local memory, use it on shared memory-enabled GPUs
float GetPeak_local(__local float* fp_PoT, int ul_TOffset, int ul_HalfSumLength, 
					   float f_MeanPower, float f_PeakScaleFactor, int ul_FftLength,__local float* f_weight) {
	// Peak power is calculated as the weighted
	// sum of all powers within ul_HalfSumLength
	// of the assumed gaussian peak.
	// The weights are given by the gaussian function itself.
	// BUG WATCH : for the f_PeakScaleFactor to work,
	// ul_HalfSumLength *must* be set to sigma.

	int i;
	float f_sum=(float)0.0f;

	// Find a weighted sum
	for (i = ul_TOffset - ul_HalfSumLength; i <= ul_TOffset + ul_HalfSumLength; i++) {
		f_sum += (fp_PoT[i*get_local_size(0)+get_local_id(0)] - f_MeanPower) * f_weight[abs(i-ul_TOffset)];
	}
	return(f_sum * f_PeakScaleFactor);
}
float GetChiSq_local(__local float* fp_PoT, int ul_FftLength,int ul_TOffset, 
				float f_PeakPower, float f_MeanPower, float* xsq_null,__local float* f_weight) {
	// We calculate our assumed gaussian powers
	// on the fly as we try to fit them to the
	// actual powers at each point along the PoT.

	float f_ChiSq = (float)0.0f, f_null_hyp=(float)0.0f;
	float rebin = 1024*1024 / ul_FftLength / 64;

	float recip_MeanPower = (float)1.0f / f_MeanPower;	
	for (int i = 0; i < 64; ++i) {				
		float f_PredictedPower = f_MeanPower + f_PeakPower * (float)f_weight[abs(i - ul_TOffset)];		
		f_PredictedPower *= recip_MeanPower;

		// ChiSq in this realm is:
		//  sum[0:i]( (observed power - expected power)^2 / expected variance )
		// The power of a signal is:
		//  power = (noise + signal)^2 = noise^2 + signal^2 + 2*noise*signal
		// With mean power normalization, noise becomes 1, leaving:
		//  power = signal^2 +or- 2*signal + 1
		float noise=2.0f*sqrt(max(f_PredictedPower,1.0f)-1.0f)+1.0f;
		float recip_noise = (float)rebin / noise;

		float PoTval = fp_PoT[i * get_local_size(0)+get_local_id(0)] * recip_MeanPower;
		f_ChiSq += (recip_noise*(PoTval - f_PredictedPower)*(PoTval - f_PredictedPower));
		f_null_hyp+= (recip_noise*(PoTval - 1.0f)*(PoTval - 1.0f));
	}

	f_ChiSq/=(float)(64.0f);
	f_null_hyp/=(float)(64.0f);

	(*xsq_null)=f_null_hyp;
	return f_ChiSq;
}

__kernel void GaussFit_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint* result_flag, __global float4* GaussFitResults,
								 __local float* p//,image2d_t gauss_cache,image2d_t null_cache
								 ) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	for(int i=0;i<(64/get_local_size(1));i++){ 
		p[(get_local_size(1)*i+get_local_id(1))*get_local_size(0)+get_local_id(0)]=PoT[tid+(get_local_size(1)*i+get_local_id(1))*(ul_FftLength)];
	}//R: fetch get_local_size(0) PoTs into local memory
	__local float l_weight[64];
	int ltid=(get_local_size(0)*get_local_id(1)+get_local_id(0));
	(ltid<64)?(l_weight[ltid]=f_weight[ltid]):false;//R: fetch weights array into local memory too
	barrier(CLK_LOCAL_MEM_FENCE);
//R: now all accesses to PoT will go into local memory.
if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_local(
		p,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,l_weight
		);
#if 1
	// worth looking at ?
		int res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_local(
		p,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,l_weight
		);
	float4 tmp;

	//R: additional check to see if score needed or not
/*
	res=(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}
*/
#if 0
	float score = calc_GaussFit_score_cached(f_ChiSq, f_null_hyp,settings->score_offset,gauss_cache,null_cache);
#else
	float score = calc_GaussFit_score_fl(f_ChiSq, f_null_hyp,settings->score_offset);
#endif
	if (((f_ChiSq <=  settings->gauss_chi_sq_thresh) && (score > best_gauss_score)) 
		|| ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()

__kernel void GaussFit_no_best_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint* result_flag, __global float4* GaussFitResults,__local float* p) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	for(int i=0;i<(64/get_local_size(1));i++){ 
		p[(get_local_size(1)*i+get_local_id(1))*get_local_size(0)+get_local_id(0)]=PoT[tid+(get_local_size(1)*i+get_local_id(1))*(ul_FftLength)];
	}//R: fetch get_local_size(0) PoTs into local memory
	__local float l_weight[64];
	int ltid=(get_local_size(0)*get_local_id(1)+get_local_id(0));
	(ltid<64)?(l_weight[ltid]=f_weight[ltid]):false;//R: fetch weights array into local memory too
	barrier(CLK_LOCAL_MEM_FENCE);
//R: now all accesses to PoT will go into local memory.
if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_local(
		p,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,l_weight
		);
#if 1
	// worth looking at ?
		int res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		//result_flag[0].w=1;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_local(
		p,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,l_weight
		);
	float4 tmp;

	if ( ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()


#else //R: ATI HD4xxx kernel (non vectorized but with pre-exit enabled)
__kernel void GaussFit_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint* result_flag, __global float4* GaussFitResults) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_fl(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 1
	int res;
	// worth looking at ?
	res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_fl(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);
	float4 tmp;

	//R: additional check to see if score needed or not
/*
	res=(f_ChiSq <=  (float4)settings->gauss_chi_sq_thresh);
	if( ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))||
		((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh)) ||
		((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
	){
		result_flag[0].z=1;
	}
*/
	float score = calc_GaussFit_score_fl(f_ChiSq, f_null_hyp,settings->score_offset);
	
	if (((f_ChiSq <=  settings->gauss_chi_sq_thresh) && (score > best_gauss_score)) 
		|| ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()


__kernel void GaussFit_no_best_kernel_cl(__global float* PoT, __global float* PoTPrefixSum,int ul_FftLength, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint* result_flag, __global float4* GaussFitResults) {
    int tid =  get_global_id(0);//R: from 0 to fftlen    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
	if(ul_TOffset>=settings->GaussTOffsetStop) return;//R: we outside of possible gaussian
	float f_null_hyp;

	int iSigma = settings->iSigma;

	float 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2_fl(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak_fl(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 1
	int res;
	// worth looking at ?
	res=(f_PeakPower  < settings->GaussPeakPowerThresh3*f_TrueMean);
	//res&=1;
	//debug[ul_TOffset * ul_FftLength + tid]=res;
	if (res) {
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq_fl(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);
	float4 tmp;
	if ( ((f_ChiSq <=  settings->GaussChiSqThresh) && (f_null_hyp >= settings->gauss_null_chi_sq_thresh))
		) {
//R: global size is power of 2 so it should be sahe to perform integer division here
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean;
		tmp.y=f_PeakPower;
		tmp.z=f_ChiSq;
		tmp.w=f_null_hyp;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + tid] = tmp;
} // End of gaussfit()



#endif

__kernel void GaussFit_kernel_PE_cl(__global float4* PoT, __global float4* PoTPrefixSum,int ul_FftLength, float best_gauss_score, 
								 __constant ocl_GaussFit_t* settings,__constant float* f_weight,
								 __global uint* result_flag, __global float4* GaussFitResults,
								 __constant uint* WeakPeaks) {
//R: each workitem handles 1 dot from 4 PoTs at once.
    int tid =  get_global_id(0);//R: from 0 to fftlen/4    
	int ul_TOffset =  get_global_id(1) + settings->GaussTOffsetStart;//R: something less than 64
//	if(ul_TOffset>=gpu_settings.GaussTOffsetStop) return;//R: we outside of possible gaussian
	if(WeakPeaks[tid]==0){
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		return;//R: preliminary exit if whole 4 PoTs don't have even weak peak
	}
	float4 f_null_hyp;

	int iSigma = settings->iSigma;

	float4 	f_TrueMean,
		f_ChiSq,
		f_PeakPower;

	// slide dynamic gaussian across the Power Of Time array

	// TrueMean is the mean power of the data set minus all power
	// out to 2 sigma from our current TOffset.	
	f_TrueMean = GetTrueMean2(
		PoTPrefixSum,
		ul_TOffset,
		2 * iSigma,
		ul_FftLength,tid
		);

	f_PeakPower = GetPeak(
		PoT,
		ul_TOffset,
		iSigma,
		f_TrueMean,
		settings->PeakScaleFactor,
		ul_FftLength,tid,f_weight
		);
#if 0
//R: cause each workitem deals with 4 PoTs preliminary exit possible only when all 4 have no peak
	// worth looking at ?
	if (f_PeakPower.x / f_TrueMean.x < settings->GaussPeakPowerThresh3 && f_PeakPower.y / f_TrueMean.y < settings->GaussPeakPowerThresh3 &&
		f_PeakPower.z / f_TrueMean.z < settings->GaussPeakPowerThresh3 && f_PeakPower.w / f_TrueMean.w < settings->GaussPeakPowerThresh3) {
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = (float4)0.0f;
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = (float4)0.0f;
		return;
	}
#endif
	// look at it - try to fit  
	f_ChiSq = GetChiSq(
		PoT,
		ul_FftLength,
		ul_TOffset,
		f_PeakPower,
		f_TrueMean,        
		&f_null_hyp,tid,f_weight
		);


	float4 score = calc_GaussFit_score(f_ChiSq, f_null_hyp,settings->score_offset);
	float4 tmp;
	if (((f_ChiSq.x <=  settings->gauss_chi_sq_thresh) && (score.x > best_gauss_score)) 
		|| ((f_ChiSq.x <=  settings->GaussChiSqThresh) && (f_null_hyp.x >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean.x;
		tmp.y=f_PeakPower.x;
		tmp.z=f_ChiSq.x;
		tmp.w=f_null_hyp.x;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)] = tmp;
	if (((f_ChiSq.y <=  settings->gauss_chi_sq_thresh) && (score.y > best_gauss_score)) 
		|| ((f_ChiSq.y <=  settings->GaussChiSqThresh) && (f_null_hyp.y >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean.y;
		tmp.y=f_PeakPower.y;
		tmp.z=f_ChiSq.y;
		tmp.w=f_null_hyp.y;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+1] = tmp;
	if (((f_ChiSq.z <=  settings->gauss_chi_sq_thresh) && (score.z > best_gauss_score)) 
		|| ((f_ChiSq.z <=  settings->GaussChiSqThresh) && (f_null_hyp.z >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean.z;
		tmp.y=f_PeakPower.z;
		tmp.z=f_ChiSq.z;
		tmp.w=f_null_hyp.z;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+2] = tmp;
	if (((f_ChiSq.w <=  settings->gauss_chi_sq_thresh) && (score.w > best_gauss_score)) 
		|| ((f_ChiSq.w <=  settings->GaussChiSqThresh) && (f_null_hyp.w >= settings->gauss_null_chi_sq_thresh))
		) {
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;//R: hot attention required to this block
		tmp.x=f_TrueMean.w;
		tmp.y=f_PeakPower.w;
		tmp.z=f_ChiSq.w;
		tmp.w=f_null_hyp.w;
	} else {
		tmp=(float4)0.0f;
	}    
		GaussFitResults[ul_TOffset * ul_FftLength + (tid<<2)+3] = tmp;
} // End of gaussfit()

__kernel __attribute__((work_group_size_hint(1, 1, 1)))
void set_mem_kernel_cl(__global float4* data){
	uint tid=get_global_id(0);
	data[tid]=(float4)(0.0f);
}

__kernel __attribute__((vec_type_hint(float4)))
void PC_find_triplets_kernel_cl(int ul_FftLength, int len_power, float triplet_thresh_base, int AdvanceBy, int PoTLen,
									  __global float4* PoT,__global uint* result_flag) {
//R: difference from original kernel: this one just doing fast precheck and relies on CPU for real triplet analysis
//R: this kernel does 4 PoT chunks at once.
//R: cause workitems can write flag in arbitrary order it's not possible to set error code actually (original CUDA code
// missed this fact and tries to return error code from kernel. That is, different variable should be used for setting 
// state flags. In this kernel next convention used:
//If whole flag is zero - there is no errors and no result loading on CPU required, kernel launch was silent.
//If result_flag.x not zero - but all other parts are zero - kernel launch was OK, some triplets found, CPU logging required.
// result_flag.y!=0 - there were too many bins above threshold, CPU reprocessing required
// result_flag.z!=0 -there were too many (>2) triplets found in single PoT chunk, CPU reprocessing required
//R: new in precheck kernel (only this flag can be setted in precheck kernel):
// result_flag.w!=0 - >2 bins above threshold - CPU analysys required
//	if(result_flag[0].w==1) return;//R: prev workitem already encountered such situation kernel should abort ASAP
	int ul_PoT = get_global_id(0);//R: 4 PoTs at once!
	int y = get_global_id(1);//R: index of offset chunk
	int fft_len4=ul_FftLength>>2;
	int TOffset = y * AdvanceBy;
	if(TOffset + len_power > PoTLen) {		
		TOffset = PoTLen - len_power;
	}
	__global float4* fp_PulsePot= PoT + ul_PoT + TOffset * (fft_len4);
	// Clear the result array
	int4 numBinsAboveThreshold=(int4)0;
	int i;
	float4 mean_power=(float4)0.f,triplet_thresh=(float4)triplet_thresh_base,pp;

	/* Get all the bins that are above the threshold, and find the power array mean value */
	for( i=0;i<len_power;i++ ) {
		mean_power += fp_PulsePot[i*fft_len4];
	}
	mean_power /= (float4)len_power;
	triplet_thresh*=mean_power;

	for( i=0;i<len_power;i++ ) {
		pp= fp_PulsePot[i*fft_len4];
		if(  pp.x>= triplet_thresh.x ) {
			numBinsAboveThreshold.x++;
		}
		if(  pp.y>= triplet_thresh.y ) {
			numBinsAboveThreshold.y++;
		}
		if(  pp.z>= triplet_thresh.z ) {
			numBinsAboveThreshold.z++;
		}
		if(  pp.w>= triplet_thresh.w ) {
			numBinsAboveThreshold.w++;
		}
	}
	if(numBinsAboveThreshold.x>2 || numBinsAboveThreshold.y>2 || numBinsAboveThreshold.z>2 || numBinsAboveThreshold.w>2){
//R: global size is power of 2 so it should be sahe to perform integer division here
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;
	}
}


__kernel __attribute__((vec_type_hint(float4)))
void PC_find_triplets_avg_kernel_cl(int ul_FftLength, int len_power, float triplet_thresh_base, int AdvanceBy, int PoTLen,
									  __global float4* PoT,__global uint* result_flag,__global float4* PulsePoT_average) {
//R: difference from original kernel: this one just doing fast precheck and relies on CPU for real triplet analysis
//R: this kernel does 4 PoT chunks at once.
//R: cause workitems can write flag in arbitrary order it's not possible to set error code actually (original CUDA code
// missed this fact and tries to return error code from kernel. That is, different variable should be used for setting 
// state flags. In this kernel next convention used:
//If whole flag is zero - there is no errors and no result loading on CPU required, kernel launch was silent.
//If result_flag.x not zero - but all other parts are zero - kernel launch was OK, some triplets found, CPU logging required.
// result_flag.y!=0 - there were too many bins above threshold, CPU reprocessing required
// result_flag.z!=0 -there were too many (>2) triplets found in single PoT chunk, CPU reprocessing required
//R: new inj precheck kernel (only this flag can be setted in precheck kernel):
// result_flag.w!=0 - >2 bins above threshold - CPU analysys required
//	if(result_flag[0].w==1) return;//R: prev workitem already encountered such situation kernel should abort ASAP
	int ul_PoT = get_global_id(0);//R: 4 PoTs at once!
	int y = get_global_id(1);//R: index of offset chunk
	int fft_len4=ul_FftLength>>2;
	int TOffset = y * AdvanceBy;
	if(TOffset + len_power > PoTLen) {		
		TOffset = PoTLen - len_power;
	}
	__global float4* fp_PulsePot= PoT + ul_PoT + TOffset * (fft_len4);
	// Clear the result array
	int4 numBinsAboveThreshold=(int4)0;
	int i;
	float4 mean_power=(float4)0.f,triplet_thresh=(float4)triplet_thresh_base,pp;

	/* Get all the bins that are above the threshold, and find the power array mean value */
	for( i=0;i<len_power;i++ ) {
		mean_power += fp_PulsePot[i*fft_len4];
	}
	mean_power /= (float4)len_power;
	PulsePoT_average[ul_PoT+y*fft_len4]=mean_power;//R: this avg will be needed later, at pulse finding
	triplet_thresh*=mean_power;

	for( i=0;i<len_power;i++ ) {
		pp= fp_PulsePot[i*fft_len4];
		if(  pp.x>= triplet_thresh.x ) {
			numBinsAboveThreshold.x++;
		}
		if(  pp.y>= triplet_thresh.y ) {
			numBinsAboveThreshold.y++;
		}
		if(  pp.z>= triplet_thresh.z ) {
			numBinsAboveThreshold.z++;
		}
		if(  pp.w>= triplet_thresh.w ) {
			numBinsAboveThreshold.w++;
		}
	}
	if(numBinsAboveThreshold.x>2 || numBinsAboveThreshold.y>2 || numBinsAboveThreshold.z>2 || numBinsAboveThreshold.w>2){
//R: global size is power of 2 so it should be sahe to perform integer division here
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;
	}
}



/*
0 1 2 3 4 5 6 7 8 9 10 - 11=di
x x x x x - new di
x x x x x x - tmp0=1+new di=6
0+6 1+7 2+8 3+9 (4+10)=stop+(stop+tmp0) - 5=di
 x   x  -new di, tmp0=1+2=3
(0+6)+(3+9) (1+7)+(4+10)
*/ 
float4 sumtop2(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		sum = i1 + i2;
		dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop3(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1) {
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	__global float4 *three = tab + tmp1 * (fft_len4);

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len4);
		float4 i1 = one[offset];
		float4 i2 = two[offset];
		float4 i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float4 sumtop4(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop5(__global float4 *tab, __global float4* dest, int di, int fft_len4, int tmp0, int tmp1, int tmp2, int tmp3) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	__global float4 *five = tab + tmp3 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		float4 i5 = five[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
typedef struct { 
	//int NumDataPoints;
	// find_triplets
	//float *power_ft;
	//float4* results_ft;
	//result_flag* result_flags_ft; 	

	// find_pulse
	//float* PulsePot_fp;		// Input data
	//float* PulsePot8_fp;	// Input data moved 8 bytes forward for coleased reads
	//float* tmp_pot_fp;		// Temporary array
	//float* best_pot_fp;		// Copy folded pots with best score
	//float* report_pot_fp;	// Copy folded pots for reporting
	//float4* results_fp;		// Additional data for reporting

	//result_find_pulse_flag* result_flags_fp;

	//const float* t_funct_cache_fp; // cached results of cudaAcc_t_funct								   								   
	float rcfg_dis_thresh;
	int PulseMax;								                                      
} PulseFind_t;
#define FOLDS_COUNT 4
#define FOLDS_START 2
float4 t_funct(int di, int num_adds, int j, int PulseMax, __global float* t_funct_cache) {
	return (float4)t_funct_cache[j * PulseMax * FOLDS_COUNT + (num_adds - FOLDS_START) * PulseMax + di];
}

__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_kernel1_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	uint y = get_global_id(1);
	uint z=get_global_id(2);//Periods unroll
	uint TOffset1 = y * AdvanceBy;
	uint TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	uint fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=get_global_size(2)) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel
void PC_find_pulse_kernel2_pass3_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,//int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,
								  __global float4* PulsePoT_average) 
{
//	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/3);//pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];//*3.f;//R: will use scaled max later
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = 3;//pass;
	//int res=0;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2;
		//int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=get_global_size(2)) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper;//, perdiv;
			int tmp0, tmp1;//, tmp2, tmp3;
			float4 scaled_avg;
			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/2);//num_adds_minus1);
			//perdiv = 2;//num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/2;//perdiv;                      // (int)period			
			dis_thresh = t_funct(di, 3/*num_adds*/, 0, PulseMax, t_funct_cache_fp) * avg;
			tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				scaled_avg=avg*3;
				//tmp_max /= 3;//num_adds;
				cur_thresh = (dis_thresh/* / 3 num_adds*/ - scaled_avg) * rcfg_dis_thresh + scaled_avg;

				float4 _snr = (tmp_max-scaled_avg)/(cur_thresh-scaled_avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score || (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					//res=1;
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 =6;// 2* num_adds;

			//	int j = 1;
			//float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di,3/* num_adds*/, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					scaled_avg=avg*num_adds_2;
					//tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh/* / num_adds_2*/ - scaled_avg) * rcfg_dis_thresh + scaled_avg;

					float4 _snr = (tmp_max-scaled_avg)/(cur_thresh-scaled_avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score || 
						(tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						//res=1;
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
	//if(res==1)result_flag[0].x = 1;
}
__kernel __attribute__((vec_type_hint(float4)))
void PC_find_pulse_partial_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,int P,int PP) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (32*y+z) * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > PP ; p-=32) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
__kernel void PC_find_pulse_partial_kernel1_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global float4* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float4* PulsePoT_average,int P,int PP) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int lastP;
		switch(num_adds) {
		case 3: lastP = P-z; break;
		case 4: lastP = P-z; break;
		case 5: lastP = P-z; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > PP ; p-=get_global_size(2)) {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5(fp_PulsePot, tmp_pot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			float4 tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2(tmp_pot, tmp_pot, di, fft_len4, tmp0);
				}
				if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y ||
					tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr.x  > best_pulse_score ||
						_snr.y  > best_pulse_score ||
						_snr.z  > best_pulse_score ||
						_snr.w  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max.x>cur_thresh.x) ||
						(tmp_max.y>cur_thresh.y) ||
						(tmp_max.z>cur_thresh.z) ||
						(tmp_max.w>cur_thresh.w) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}


float4 sumtop3f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1) {
	float4 sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * fft_len4;
	__global float4 *three = tab + tmp1 * (fft_len4);

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * (fft_len4)];
		float4 i2 = two[i * (fft_len4)];
		float4 i3 = three[i * (fft_len4)];
		sum = i1 + i2;
		sum += i3;
		//dest[i * (fft_len4)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop4f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1, int tmp2) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		//dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float4 sumtop5f(__global float4 *tab, int di, int fft_len4, int tmp0, int tmp1, int tmp2, int tmp3) {
	float4 sum, tmax;
	int   i;
	__global float4 *one = tab;
	__global float4 *two = tab + tmp0 * (fft_len4);
	__global float4 *three = tab + tmp1 * (fft_len4);
	__global float4 *four = tab + tmp2 * (fft_len4);
	__global float4 *five = tab + tmp3 * (fft_len4);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		float4 i1 = one[i * fft_len4];
		float4 i2 = two[i * fft_len4];
		float4 i3 = three[i * fft_len4];
		float4 i4 = four[i * fft_len4];
		float4 i5 = five[i * fft_len4];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		//dest[i * fft_len4] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
__kernel __attribute__((vec_type_hint(float4)))
//R: TODO: cause kernel is small need to check 
//1) if reading flag is needed- reading global memory is costly
//2) if unroll over periods (z-adis) is needed - each unrolled workitem requires separate average read from global memory
void PC_find_pulse_f_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float4* PoT,__global uint* result_flag,
								  __global float4* PulsePoT_average) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	int z=get_global_id(2);
	int firstP = (PulsePotLen*1)/2;
	int p = (PulsePotLen*2)/3-z;
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);

	int TOffset1 = y * AdvanceBy;
	//int TOffset2 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float4* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len4);
	//__global float4* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len4);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float4 avg;//,maxp=0;//, snr=0, fthresh=0;
	float4 tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len4];
	uint need_cpu_reprocess=0;
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		switch(num_adds) {
		case 4: p = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; 
			break;
		case 5: p = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; 
			break;
		}
		if(p<=firstP) return;//R: return should be safe cause first iteration has biggest number of periods involved
		int num_adds_minus1 = num_adds - 1;
		//int p = lastP;
		/*for (p = lastP ; p > firstP ; p-=)*/ {
			float4 cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3f(fp_PulsePot, di, fft_len4, tmp0, tmp1);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4f(fp_PulsePot, di, fft_len4, tmp0, tmp1, tmp2);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5f(fp_PulsePot, di, fft_len4, tmp0, tmp1, tmp2, tmp3);
				break;
			}

			if (tmp_max.x>dis_thresh.x || tmp_max.y>dis_thresh.y || 
				tmp_max.z>dis_thresh.z || tmp_max.w>dis_thresh.w) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float4 _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr.x  > best_pulse_score ||
					_snr.y  > best_pulse_score ||
					_snr.z  > best_pulse_score ||
					_snr.w  > best_pulse_score ) {
					need_cpu_reprocess = 1;					
				}
				if( (tmp_max.x>cur_thresh.x) ||
					(tmp_max.y>cur_thresh.y) ||
					(tmp_max.z>cur_thresh.z) ||
					(tmp_max.w>cur_thresh.w) ) {
					need_cpu_reprocess = 1;//return;
				}
			}
		} // for (p = lastP
	} // for(num_adds =
	//R: global size is power of 2 so it should be safe to perform integer division here
	if(need_cpu_reprocess){	
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;
	}

}




#if USE_OPENCL_NV || USE_OPENCL_HD5xxx
float sumtop2_local(__local float *tab, __local float* dest, int di,int tmp0) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__local float *one = tab;
	__local float *two = tab + tmp0*get_local_size(0);

	for (i = 0; i < di; i++) {
		int offset=i*get_local_size(0);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop3_local( __global float *tab, __local float* dest, int di, int tmp0, int tmp1,int fftlen) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_local(__global float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2,int fftlen) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;
	__global float *four = tab + tmp2*fftlen;
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop5_local(__global float *tab, __local float* dest, int di, int tmp0, int tmp1, int tmp2, int tmp3,int fftlen) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0*fftlen;
	__global float *three = tab + tmp1*fftlen;
	__global float *four = tab + tmp2*fftlen;
	__global float *five = tab + tmp3*fftlen;
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i*fftlen;
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[i*get_local_size(0)] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float t_funct_f(int di, int num_adds, int j, int PulseMax, __global float* t_funct_cache) {
	return t_funct_cache[j * PulseMax * FOLDS_COUNT + (num_adds - FOLDS_START) * PulseMax + di];
}
//R: this kernel should be used for small power arrays that fit in local memory comletely
__kernel void PC_find_pulse_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__local float* tmp_PoT,__global uint* result_flag,
								  __global float* PulsePoT_average) 
{
	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	int y = get_global_id(1);
	int z=get_local_id(0);
	int TOffset1 = y * AdvanceBy;
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	int need_cpu_reprocess=0;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__local float* tmp_pot		= tmp_PoT +z;
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	//int num_adds = pass;
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p--) {
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1,fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2,fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_local(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2, tmp3,fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score || (tmp_max>cur_thresh)) {
					need_cpu_reprocess = 1;break;					
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2_local(tmp_pot, tmp_pot, di,tmp0);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score || (tmp_max>cur_thresh)) {
						need_cpu_reprocess = 1;break;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
		//R: global size is power of 2 so it should be safe to perform integer division here
	if(need_cpu_reprocess){	
		int result_coordinate=(get_global_size(0)>RESULT_SIZE)?
			((RESULT_SIZE*get_global_id(0))/get_global_size(0)):get_global_id(0);
		result_flag[result_coordinate]=1;
	}

}
float sumtop3_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int fft_len) {
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;
	__global float *three = tab + tmp1 * (fft_len);

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		sum = i1 + i2;
		sum += i3;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop4_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int tmp2, int fft_len) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0 * (fft_len);
	__global float *three = tab + tmp1 * (fft_len);
	__global float *four = tab + tmp2 * (fft_len);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;		
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}

float sumtop5_fl(__global float *tab, __global float* dest, int di, int tmp0, int tmp1, int tmp2, int tmp3, int fft_len) {
	float sum, tmax;
	int   i;
	__global float *one = tab;
	__global float *two = tab + tmp0 * (fft_len);
	__global float *three = tab + tmp1 * (fft_len);
	__global float *four = tab + tmp2 * (fft_len);
	__global float *five = tab + tmp3 * (fft_len);
	tmax = 0.0f;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		float i3 = three[offset];
		float i4 = four[offset];
		float i5 = five[offset];
		sum = i1 + i2;
		sum += i3;
		sum += i4;
		sum += i5;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}	
	return tmax;
}
float sumtop2_fl(__global float *tab, __global float* dest, int di, int tmp0, int fft_len) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop2_semi_local(__global float *tab, __local float* dest, int di, int tmp0, int fft_len,int z) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__global float *one = tab;
	__global float *two = tab + tmp0 * fft_len;

	for (i = 0; i < di; i++) {
		int offset=i * (fft_len);
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[i*z] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
float sumtop2_local2(__local float *tab, __local float* dest, int di,int tmp0,int z) {
//R: this function working with 4 different frequency bins at once as all other pulse finding functions do
	float sum, tmax;
	int   i;
	tmax = 0.0f;
	__local float *one = tab;
	__local float *two = tab + tmp0*z;

	for (i = 0; i < di; i++) {
		int offset=i*z;
		float i1 = one[offset];
		float i2 = two[offset];
		sum = i1 + i2;
		dest[offset] = sum;
		tmax = max(tmax, sum);		
	}
	return tmax;
}
__kernel void PC_find_pulse_semi_local_kernel_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__global float* tmp_PoT,__global uint4* result_flag,
								  __global float* PulsePoT_average,__local float* local_tmp) 
{
//R: X-axis - different FFT bins, Y-axis - different PulsePoTs inside same FFT bin
//R: Z-axis - unroll through periods

	//if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0);
	if (ul_PoT < 1) return; // Original find_pulse, omits first element
	int y = get_global_id(1);
	int z=get_global_id(2);
	int slz=get_local_size(0)*get_local_size(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 =(get_global_size(2)*y+get_global_id(2)) * (PulsePotLen/3);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;

	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__global float* tmp_pot=tmp_PoT + ul_PoT + TOffset2 * (fft_len);		
	__local float* ltmp=local_tmp+get_local_id(0)+get_local_id(2)*get_local_size(0);

	int di;
	
	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=get_global_size(2)) {
			int local_only=0;
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1,fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2,fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, tmp2, tmp3,fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float4)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				if(local_only)//R: both arrays in local memory already
				{
					tmp_max = sumtop2_local2(ltmp, ltmp, di,tmp0,slz);
				}else if(di<=60)//R:new array will fit inside local memory
				{
					local_only=1;
					tmp_max = sumtop2_semi_local(tmp_pot, ltmp, di,tmp0,fft_len,slz);
				}else
				{
					tmp_max = sumtop2_fl(tmp_pot, tmp_pot, di,tmp0,fft_len);
				}
				if (tmp_max>dis_thresh) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
#if USE_OPENCL_NV
__kernel void PC_find_pulse_kernel1_fl_cl(float best_pulse_score, int PulsePotLen, int AdvanceBy,
								  int fft_len, int ndivs,int pass,
								  __global PulseFind_t* settings,__global float* t_funct_cache_fp,
								  __global float* PoT,__global float* tmp_PoT,__global uint4* result_flag,int offset,
								  __global float* PulsePoT_average) 
{
	if(result_flag[0].x == 1) return;//R: if CPU processing already needed - abort kernel ASAP
	const int PoTLen = 1024*1024/fft_len;
	int ul_PoT = get_global_id(0)+offset;
	int y = get_global_id(1);
	int z=get_global_id(2);
	int TOffset1 = y * AdvanceBy;
	int TOffset2 = (get_global_size(2)*y+z) * (PulsePotLen/pass);
	float rcfg_dis_thresh=settings->rcfg_dis_thresh;
	int PulseMax=settings->PulseMax;
	//int fft_len4=fft_len>>2;

	//if (ul_PoT < 1) return; // Original find_pulse, omits first element
	//if (ul_PoT >= fft_len) return;
	if(TOffset1 + PulsePotLen > PoTLen) {		
		TOffset1 = PoTLen - PulsePotLen;
	}

	__global float* fp_PulsePot	= PoT + ul_PoT + TOffset1 * (fft_len);
	__global float* tmp_pot		= tmp_PoT + ul_PoT + TOffset2 * (fft_len);
	//float* best_pot		= cudaAcc_PulseFind_settings.best_pot_fp	+ ul_PoT + TOffset2 * fft_len;
	//float* report_pot	= cudaAcc_PulseFind_settings.report_pot_fp	+ ul_PoT + TOffset2 * fft_len;

	int di;

	float avg;//,maxp=0;//, snr=0, fthresh=0;
	float tmp_max;
	avg=PulsePoT_average[ul_PoT+y*fft_len];
	//  Periods from PulsePotLen/3 to PulsePotLen/4, and power of 2 fractions of.
	//   then (len/4 to len/5) and finally (len/5 to len/6)
	//	
	int num_adds = pass;
	//for(int num_adds = 3; num_adds <= 5; num_adds++) 
    {
		int firstP, lastP;
		switch(num_adds) {
		case 3: lastP = (PulsePotLen*2)/3-z;  firstP = (PulsePotLen*1)/2; break;
		case 4: lastP = (PulsePotLen*3)/4-z;  firstP = (PulsePotLen*3)/5; break;
		case 5: lastP = (PulsePotLen*4)/5-z;  firstP = (PulsePotLen*4)/6; break;
		}
		int num_adds_minus1 = num_adds - 1;
		int p = lastP;
		for (p = lastP ; p > firstP ; p-=get_global_size(2)) {
			float cur_thresh, dis_thresh;
			int /*tabofst, */mper, perdiv;
			int tmp0, tmp1, tmp2, tmp3;

			//tabofst = ndivs*3+2-num_adds;
			mper = p * (12/num_adds_minus1);
			perdiv = num_adds_minus1;
			tmp0 = (int)((mper + 6)/12);             // round(period)
			tmp1 = (int)((mper * 2 + 6)/12);         // round(period*2)
			di = (int)p/perdiv;                      // (int)period			
			//dis_thresh = cudaAcc_t_funct(di, num_adds)*avg;
			dis_thresh = t_funct_f(di, num_adds, 0, PulseMax, t_funct_cache_fp) * avg;

			switch(num_adds) {
			case 3:
				tmp_max = sumtop3_fl(fp_PulsePot, tmp_pot, di, tmp0, tmp1, fft_len);
				break;
			case 4:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp_max = sumtop4_fl(fp_PulsePot, tmp_pot, di,tmp0, tmp1, tmp2, fft_len);
				break;
			case 5:
				tmp2 = (int)((mper * 3 + 6)/12);     // round(period*3)
				tmp3 = (int)((mper * 4 + 6)/12);     // round(period*4)
				tmp_max = sumtop5_fl(fp_PulsePot, tmp_pot, di,tmp0, tmp1, tmp2, tmp3, fft_len);
				break;
			}

			if (tmp_max>dis_thresh) {
				// unscale for reporting
				tmp_max /= num_adds;
				cur_thresh = (dis_thresh / num_adds - avg) * rcfg_dis_thresh + avg;

				float _snr = (tmp_max-avg)/(cur_thresh-avg);
				if (_snr  > best_pulse_score ) {
					result_flag[0].x = 1;//return;					
				}
				if( (tmp_max>cur_thresh) ) {
					result_flag[0].x = 1;//return;
				}
			}

			int num_adds_2 = 2* num_adds;

			//	int j = 1;
			//float tmp_max2=(float)0.f;
			for (int j = 1; j < ndivs ; j++) 
			{
				//perdiv *=2;
				tmp0 = di & 1;
				di >>= 1;
				tmp0 += di;
				dis_thresh = t_funct_f(di, num_adds, j,PulseMax,t_funct_cache_fp) * avg;
				{
					tmp_max = sumtop2_fl(tmp_pot, tmp_pot, di, tmp0, fft_len);
				}
				if (tmp_max>dis_thresh ) {
					// unscale for reporting
					tmp_max /= num_adds_2;
					cur_thresh = (dis_thresh / num_adds_2 - avg) * rcfg_dis_thresh + avg;

					float _snr = (tmp_max-avg)/(cur_thresh-avg);
					if (_snr  > best_pulse_score) {
						result_flag[0].x = 1;return;
					}
					if ((tmp_max>cur_thresh) ) {
						result_flag[0].x = 1;//return;
					}
				}

				num_adds_2 *=2;
			}  // for (j = 1; j < ndivs
		} // for (p = lastP
	} // for(num_adds =
}
#endif
#endif

#if SETI7
__kernel void RepackInput_kernel_cl( __global float* PowerSpectrum, __global float2* dct_In) 
{
	//R: taking Jason's CUDA kernel as base and extend to 2D matrix to process whole icfft at once 
	int sidx = get_global_id(0);
	int y=get_global_id(1);
	int nDestPoints = (get_global_size(0))*4;
	int didx1 = sidx*2; 
	int didx2 = (nDestPoints-2)-didx1+nDestPoints*y;
	didx1+=nDestPoints*y;

	float a = PowerSpectrum[sidx+y*get_global_size(0)];

	dct_In[didx1] =(float2)(0.0f,0.f);
	dct_In[didx1+1] =(float2)(a,0.0f);

	dct_In[didx2] =(float2)(0.0f,0.0f);
	dct_In[didx2+1] =(float2)(a,0.0f);
}
#if 0 //USE_OPENCL_NV || USE_OPENCL_HD5xxx //R: working worse still
__kernel void FindAutoCorrelation_kernel1_cl(__global float2* AutoCorrelation,  __global float2* result){
	__local float* PartialSumMaxIdx[256*3];
	const int tid=get_local_id(0);
	const int y=get_global_id(1)*(4*128*1024);
	float partial_sum=0.0f;
	float partial_max=0.0f;
	float idx=0.0f;
	__global float2* auto_corr=AutoCorrelation+y;
	for(int i=tid*256;i<(tid+1)*256;i++){
		float tmp=auto_corr[i].x;
		tmp*=tmp;
		partial_sum+=tmp;
		if(i!=0 && tmp>partial_max){
			partial_max=tmp;
			idx=(float)i;
		}
	}
	int did=3*tid;
	PartialSumMaxIdx[did]=partial_sum;
	PartialSumMaxIdx[did+1]=partial_max;
	PartialSumMaxIdx[did+2]=idx;
	if(tid==0){
		float full_max=0.0f;
		float idx=0.0f;
		float Sum=0.0f;
		for(int i=0;i<256;i++){
			Sum+=PartialSumMaxIdx[3*i];
			if(PartialSumMaxIdx[i*3+1]>full_max){
				full_max=PartialSumMaxIdx[i*3+1];
				idx=PartialSumMaxIdx[i*3+2];
			}
		}
		result[get_global_id(1)].x=full_max*(128.0f*1024.0f)/Sum;//peak value
		result[get_global_id(1)].y=idx;//bin
	}

}
#else
__kernel void FindAutoCorrelation_reduce0_kernel1_cl(__global float2* AutoCorrelation, __global float* PartialSumMaxIdx){
	const int tid=get_global_id(0);
	const int y=get_global_id(1)*(4*128*1024);
	float partial_sum=0.0f;
	float partial_max=0.0f;
	float idx=0.0f;
	__global float2* auto_corr=AutoCorrelation+y;
	for(int i=tid*256;i<(tid+1)*256;i++){
		float tmp=auto_corr[i].x;
		tmp*=tmp;
		partial_sum+=tmp;
		if(i!=0 && tmp>partial_max){
			partial_max=tmp;
			idx=(float)i;
		}
	}
	int did=3*(256*get_global_id(1)+tid);
	PartialSumMaxIdx[did]=partial_sum;
	PartialSumMaxIdx[did+1]=partial_max;
	PartialSumMaxIdx[did+2]=idx;
}
__kernel void FindAutoCorrelation_reduce1_kernel_cl(__global float* PartialSumMaxIdx, __global float2* result){
	const int sid=256*get_global_id(0);//only 8 workitems here
	float full_max=0.0f;
	float idx=0.0f;
	float Sum=0.0f;
	for(int i=0;i<256;i++){
		Sum+=PartialSumMaxIdx[3*(sid+i)];
		float tmp=PartialSumMaxIdx[(sid+i)*3+1];
		if(tmp>full_max){
			full_max=tmp;
			idx=PartialSumMaxIdx[(sid+i)*3+2];
		}
	}
	result[get_global_id(0)].x=full_max*(128.0f*1024.0f)/Sum;//peak value
	result[get_global_id(0)].y=idx;//bin

}
#endif
#endif
